@@ -30,6 +30,8 @@
 #include "config.h"
 #endif
 
+#include <stdint.h>
+
 #include "xf86.h"
 #include "xf86_OSproc.h"
 #include "xf86Resources.h"
@@ -44,8 +46,23 @@
 #include "tfp410.h"
 #include "tfp410_reg.h"
 
+typedef struct _TFP410SaveRec {
+    uint8_t ctl1;
+    uint8_t ctl2;
+} TFP410SaveRec;
+
+typedef struct {
+    I2CDevRec d;
+    Bool quiet;
+
+    TFP410SaveRec SavedReg;
+    TFP410SaveRec ModeReg;
+} TFP410Rec, *TFP410Ptr;
+
+#define TFPPTR(d) ((TFP410Ptr)(d->DriverPrivate.ptr))
+
 static Bool
-tfp410ReadByte(TFP410Ptr tfp, int addr, CARD8 *ch)
+tfp410ReadByte(TFP410Ptr tfp, int addr, uint8_t *ch)
 {
     if (!xf86I2CReadByte(&(tfp->d), addr, ch)) {
 	if (!tfp->quiet) {
@@ -59,7 +76,7 @@ tfp410ReadByte(TFP410Ptr tfp, int addr, CARD8 *ch)
 }
 
 static Bool
-tfp410WriteByte(TFP410Ptr tfp, int addr, CARD8 ch)
+tfp410WriteByte(TFP410Ptr tfp, int addr, uint8_t ch)
 {
     if (!xf86I2CWriteByte(&(tfp->d), addr, ch)) {
 	if (!tfp->quiet) {
@@ -191,7 +208,7 @@ static void
 tfp410_dump_regs(I2CDevPtr d)
 {
     TFP410Ptr tfp = TFPPTR(d);
-    CARD8 val, val2;
+    uint8_t val, val2;
 
     tfp410ReadByte(tfp, TFP410_REV, &val);
     xf86DrvMsg(tfp->d.pI2CBus->scrnIndex, X_INFO,
@@ -259,7 +276,7 @@ tfp410_restore(I2CDevPtr d)
     tfp410WriteByte(tfp, TFP410_CTL_1, tfp->SavedReg.ctl1);
 }
 
-I830I2CVidOutputRec TFP410VidOutput = {
+_X_EXPORT I830I2CVidOutputRec TFP410VidOutput = {
     .init = tfp410_init,
     .detect = tfp410_detect,
     .mode_valid = tfp410_mode_valid,
